import numpy as np
import matplotlib
import matplotlib.pyplot as plt
from matplotlib.collections import PatchCollection
import matplotlib.path as mpath
import matplotlib.patches as mpatches
import matplotlib.lines as mlines
import matplotlib.text as text
import matplotlib.font_manager as fm 
import pylab as pl
import pandas as pd
from textwrap import wrap

from .chem import *

class ChemView(ChemImage):
  def __init__(self,*args,ax=None,z0=1,fs=1.2,xyoff=[0,0],lw=1.0,dbg=False,
               **kwargs):
    ChemImage.__init__(self,*args,**kwargs)
    self.dbg=dbg
    self.HA = dict(middle='center',end='right',start='left')
    self.ax = ax
    self.z0 = z0
    self.fs = fs
    self.lw = lw
    self.xyoff=xyoff

  def __call__(self,smi,x0=0.0,y0=0.0,h=None,w=None,scale=1.0):
    try:
      Img = self.smi2img(smi)
    except:
      pass
    else:
      w = w if w else self._molsize[0]
      h = h if h else self._molsize[1]
      hw,hh = 0.5 * w, 0.5 * h
      self.ax.imshow(Img,extent=(x0-hw,x0+hw,y0-hw,y0+hw))
    
  
class ChemJSView(ChemJson):
  def __init__(self,*args,ax=None,z0=1,fs=1.2,xyoff=[0,0],lw=1.0,dbg=False,
               **kwargs):
    ChemJson.__init__(self,*args,**kwargs)
    self.dbg=dbg
    self.HA = dict(middle='center',end='right',start='left')
    self.ax = ax
    self.z0 = z0
    self.fs = fs
    self.lw = lw
    self.xyoff=xyoff
    
  def __call__(self,smi,x0=0.0,y0=0.0,ball_stick=False,atom_r=5.0):
    J = self.smi2json(smi) if type(smi)==str else smi
      
    if 'text' in J:
      if ball_stick:
        self.drawBS(J['text'],x0=x0,y0=y0,atom_r=atom_r)
      else:
        self.drawText(J['text'],x0=x0,y0=y0)
    if 'path' in J:
      self.drawPath(J['path'],x0=x0,y0=y0)
      
  def drawText(self,G,x0=0.0,y0=0.0):
    xyoff=self.xyoff
    for g in G:
      if self.dbg: 
        print("Text: {x:.1f} {y:.1f} {text} {fill} {dominant_baseline} {text_anchor}".format(**g))
      horiz_anchor = self.HA[g['text_anchor']]
      txt = text.Text(g['x']+x0+xyoff[0],g['y']+y0+xyoff[1],g['text'],color=g['fill'],zorder=self.z0,
                        ha = horiz_anchor, va = 'center',
                        fontproperties=fm.FontProperties(size=g['font-size']*self.fs,
                                                         weight=g['font-weight'],
                                                         style=g['font-style'],
                                                         family=g['font-family']))
      self.ax.add_artist(txt)

  def drawPath(self,G,x0=0.0,y0=0.0):
    xyoff=self.xyoff
    for g in G:
      L = pd.DataFrame(g['lines'])
      L.loc[:,'x'] = L.x.astype(np.float)+x0
      L.loc[:,'y'] = L.y.astype(np.float)+y0

      poly = mpatches.Polygon(L[['x','y']],lw=float(g['stroke-width'])*self.lw,
                              facecolor=g['fill'],
                              edgecolor=g['stroke'],
                              zorder=self.z0,fill=g['fill']!='none',visible=True)
      pl.gca().add_patch(poly)
    
  def drawBS(self,G,x0=0.0,y0=0.0,atom_r=5):
    xyoff=self.xyoff
    for g in G:
      if self.dbg: 
        print("Text: {x:.1f} {y:.1f} {text} {fill} {dominant_baseline} {text_anchor}".format(**g))
      atoms = mpatches.Ellipse([x0+g['x']+xyoff[0],y0+g['y']+xyoff[1]],atom_r,atom_r,
                               color=g['fill'],zorder=self.z0+1)
      self.ax.add_artist(atoms)

    
class ChemJsonBS:
  def __init__(self,*args,dbg=False,**kwargs):
    ChemSVG.__init__(*args,**kwargs)    
    self.dbg=dbg
    self.HA = dict(middle='center',end='right',start='left')
    
  def draw(self,Json,ax,xmax=100,ymax=100,
           x0=0.0,y0=0.0,z0=1,lw=1.0,atom_r=2,
           fs=1.2,xyoff=[2,-5],text_align=dict(ha='center',va='top')):
    if 'text' in Json:
      for g in Json['text']:
        if self.dbg: 
          print("Text: {x:.1f} {y:.1f} {text} {fill} {dominant_baseline} {text_anchor}".format(**g))
        #horiz_anchor = self.HA[g['text_anchor']]
        atoms = mpatches.Ellipse([x0+g['x']+xyoff[0],y0+g['y']+xyoff[1]],atom_r,atom_r,
                                 color=g['fill'],zorder=z0+1)
        ax.add_artist(atoms)

    if 'path' in Json:
        for g in Json['path']:
            #P1 = np.array([map(float,p[1:].split(',')) for p in g['points'].split(' ') if len(p)>1])
            #P1[:,0] += x0
            #P1[:,1] += y0
            L = pd.DataFrame(g['lines'])
            L.loc[:,'x'] = L.x.astype(np.float)+x0
            L.loc[:,'y'] = L.y.astype(np.float)+y0
            
            poly = mpatches.Polygon(L[['x','y']],lw=float(g['stroke-width'])*lw,
                                    facecolor=g['fill'],
                                    edgecolor=g['stroke'],
                                    zorder=z0,fill=g['fill']!='none',visible=True)
            pl.gca().add_patch(poly)

    
    


# $Id: ixml.py 9 2012-05-18 18:04:29Z ishah $

import sys
import string
from lxml import etree
from lxml.etree import XMLParser
from lxml.etree import ElementTree
#from xml.parsers import *
from xml.parsers import expat
from xml.etree.ElementTree import *
import re
from copy import *
from types import *


class Element:
    'A parsed XML element'
    def __init__(self,name,attributes):
        'Element constructor'
        # The element's tag name
        self.name = name
        # The element's attribute dictionary
        self.attributes = attributes
        # The element's cdata
        self.cdata = ''
        # The element's child element list (sequence)
        self.children = []
        
    def AddChild(self,element):
        'Add a reference to a child element'
        self.children.append(element)
        
    def getAttribute(self,key):
        'Get an attribute value'
        return self.attributes.get(key)
    
    def getData(self):
        'Get the cdata'
        return self.cdata
        
    def getElements(self,name=''):
        'Get a list of child elements'
        #If no tag name is specified, return the all children
        if not name:
            return self.children
        else:
            # else return only those children with a matching tag name
            elements = []
            for element in self.children:
                if element.name == name:
                    elements.append(element)
            return elements

    def searchElement(self,name=None,elems=[]):
        if name==None: return
        for element in self.children:
            if element.name == name:
                elems.append(element)
            else:
                element.searchElement(name,elems)
        
class Xml2Obj:
    'XML to Object'
    def __init__(self):
        self.root = None
        self.nodeStack = []
        
    def StartElement(self,name,attributes):
        'SAX start element even handler'
        # Instantiate an Element object
        element = Element(name.encode(),attributes)
        
        # Push element onto the stack and make it a child of parent
        if len(self.nodeStack) > 0:
            parent = self.nodeStack[-1]
            parent.AddChild(element)
        else:
            self.root = element
        self.nodeStack.append(element)
        
    def EndElement(self,name):
        'SAX end element event handler'
        self.nodeStack = self.nodeStack[:-1]

    def CharacterData(self,data):
        'SAX character data event handler'
        if data.strip():
            #data = data.encode()
            element = self.nodeStack[-1]
            element.cdata += data
            return

    def Parse(self,text=None,filename=None):
        # Create a SAX parser
        Parser = expat.ParserCreate()

        # SAX event handlers
        Parser.StartElementHandler = self.StartElement
        Parser.EndElementHandler = self.EndElement
        Parser.CharacterDataHandler = self.CharacterData

        # Parse the XML File
        if filename != None:
            ParserStatus = Parser.Parse(open(filename,'r').read(), 1)
        elif text != None:
            ParserStatus = Parser.Parse(text, 1)                 
        return self.root
    

            
def elemDict(elem):
    E={}
    def my_decode(x):
        try:
            y=x.decode('ascii')
        except:
            y=x

        return y

    if len(elem.attributes) >0:
        for k,v in elem.attributes.items():
            if v != '':
                E[my_decode(k)]=my_decode(v)

    if elem.cdata != '':
        #print elem.cdata
        E['cdata']=my_decode(elem.cdata)

    for e in elem.children:
        #print e.name
        k = my_decode(e.name)
        if k not in E:
            E[k]=[elemDict(e)]
        else:
            if isinstance(E[k],list):
                E[k].append(elemDict(e))
            else:
                old=E[k]
                E[k]=[old,elemDict(e)]
    return E


def cleanElemDict(E,tostr=True):
    Out = {}
    def toStr(a):
        if tostr: return str(a)
        else: return a
    
    if type(E)==dict:
        for k1,v1 in E.items():
            if k1=='cdata':
                return v1
            else:
                Out[str(k1)]=cleanElemDict(v1)
        
    elif type(E)==list:
        if len(E)==1:
            return cleanElemDict(E[0])
        else:
            return [cleanElemDict(e) for e in E]

    else:
        return E
    
    return Out



from matplotlib.patches import Ellipse, Circle, FancyArrowPatch,FancyBboxPatch
from textwrap import wrap
import pandas as pd
from rdkit import Chem
import io 
from rdkit.Chem import rdDepictor
from rdkit.Chem.Draw import rdMolDraw2D, MolToImage
from rdkit.Chem.rdDepictor import GenerateDepictionMatching2DStructure
from rdkit.Chem.Draw.MolDrawing import DrawingOptions
import re

from genra.utl.svg import *

class ChemBase:
  def __init__(self,*args,kekulize=True,wedgeBonds=True, template=None, bondWidth=2,
               molsize=(200,200), **kwargs):
    self._kekulize = kekulize
    self._wedgeBonds= wedgeBonds
    self._template = template
    self._molsize  = molsize
    self._bondWidth= bondWidth

    dopt = DrawingOptions()
    dopt.bondLineWidth=bondWidth
    dopt.bondWidth = bondWidth
    dopt.colorBonds=False
    dopt.kekulize=kekulize
    self._dopt=dopt
    
  def __call__(self,smi,**kwargs):
    return self.smi2mol(smi,**kwargs)
  
  def smi2mol(self,smi):
    molSize=self._molsize
    
    mol=Chem.MolFromSmiles(smi)
    mc = Chem.Mol(mol.ToBinary())
    if self._kekulize:
      try:
          Chem.Kekulize(mc)
      except:
          mc = Chem.Mol(mol.ToBinary())
    if not mc.GetNumConformers():
        rdDepictor.Compute2DCoords(mc)
 
    Chem.SanitizeMol(mol,sanitizeOps=Chem.SanitizeFlags.SANITIZE_ALL^\
                  Chem.SanitizeFlags.SANITIZE_KEKULIZE^\
                  Chem.SanitizeFlags.SANITIZE_SETAROMATICITY)

    if self._template:
      GenerateDepictionMatching2DStructure(mol,self._template,acceptFailure=True)
    
    return mol

class ChemImage(ChemBase):
  def __init__(self,*args,**kwargs):
      super().__init__(self,*args,**kwargs)

  def __call__(self,smi,**kwargs):
    return self.smi2img(smi,**kwargs)
  
  def smi2img(self,smi,**kwargs):
      mol = self.smi2mol(smi,**kwargs)
      return MolToImage(mol,kekulize=self._kekulize,bondLineWidth=self._bondWidth,
                        wedgebonds=self._wedgeBonds,options=self._dopt)
      
class ChemSVG(ChemBase):
  
  def __init__(self,*args,**kwargs):
      ChemBase.__init__(self,*args,**kwargs)
      
  def __call__(self,smi,**kwargs):
    return self.smi2svg(smi,**kwargs)
  
  def smi2svg(self,smi,opacity=None):
    molSize=self._molsize    
    mol = self.smi2mol(smi)
    mc = Chem.Mol(mol.ToBinary())    
    drawer = rdMolDraw2D.MolDraw2DSVG(molSize[0],molSize[1])
    drawer.DrawMolecule(mc)
    drawer.FinishDrawing()
    svg = drawer.GetDrawingText()
    
    if opacity:
      svg = re.sub("rect style='opacity:[^\;]+;",
                   "rect style='opacity:{};".format(opacity),svg)
    return svg
  
class ChemJson(ChemSVG,Svg2Json):
  def __init__(self,*args,**kwargs):
      ChemSVG.__init__(self,*args,**kwargs)
      Svg2Json.__init__(self,*args,**kwargs)
  
  def __call__(self,smi,**kwargs):
    return self.smi2json(smi,**kwargs)
  
  def smi2json(self,smi,**kwargs):
    svg = self.smi2svg(smi,**kwargs)
    return self.read_svg(svg)
    
  


"""
GenRAPredValue 

Adapted from sklearn.neighbors.KNeighborsRegressor

"""

# Authors: Imran Shah (shah.imran@epa.gov)

import numpy as np
import sklearn
from sklearn.neighbors import KNeighborsRegressor
from sklearn.neighbors._base import _check_weights, _get_weights
from sklearn.utils.validation import check_X_y, check_array, check_is_fitted,_is_arraylike, _num_samples
from sklearn.utils.multiclass import unique_labels
from scipy import stats
from sklearn.utils.extmath import weighted_mode

import warnings

class GenRAPredValue(KNeighborsRegressor):
    """GenRA Value Prediction based on k-nearest neighbors.

    The target is predicted by local interpolation of the targets
    associated of the nearest neighbors in the training set.

    Parameters
    ----------
    n_neighbors : int, optional (default = 5)
        Number of neighbors to use by default for :meth:`kneighbors` queries.


    algorithm : {'auto', 'ball_tree', 'kd_tree', 'brute'}, optional
        Algorithm used to compute the nearest neighbors:

        - 'ball_tree' will use :class:`BallTree`
        - 'kd_tree' will use :class:`KDTree`
        - 'brute' will use a brute-force search.
        - 'auto' will attempt to decide the most appropriate algorithm
          based on the values passed to :meth:`fit` method.

        Note: fitting on sparse input will override the setting of
        this parameter, using brute force.

    leaf_size : int, optional (default = 30)
        Leaf size passed to BallTree or KDTree.  This can affect the
        speed of the construction and query, as well as the memory
        required to store the tree.  The optimal value depends on the
        nature of the problem.

    p : integer, optional (default = 2)
        Power parameter for the Minkowski metric. When p = 1, this is
        equivalent to using manhattan_distance (l1), and euclidean_distance
        (l2) for p = 2. For arbitrary p, minkowski_distance (l_p) is used.

    metric : string or callable, default 'minkowski'
        the distance metric to use for the tree.  The default metric is
        minkowski, and with p=2 is equivalent to the standard Euclidean
        metric. See the documentation of the DistanceMetric class for a
        list of available metrics.
        If metric is "precomputed", X is assumed to be a distance matrix and
        must be square during fit. X may be a :term:`Glossary <sparse graph>`,
        in which case only "nonzero" elements may be considered neighbors.

    metric_params : dict, optional (default = None)
        Additional keyword arguments for the metric function.

    n_jobs : int or None, optional (default=None)
        The number of parallel jobs to run for neighbors search.
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
        ``-1`` means using all processors. See :term:`Glossary <n_jobs>`
        for more details.
        Doesn't affect :meth:`fit` method.


    """

    def __init__(self, n_neighbors=5, 
                 algorithm='auto', leaf_size=30,
                 p=2, metric='minkowski', 
                 metric_params=None, n_jobs=None,
                 **kwargs):
        super().__init__(
              n_neighbors=n_neighbors,
              algorithm=algorithm,
              leaf_size=leaf_size, metric=metric, p=p,
              metric_params=metric_params, n_jobs=n_jobs, 
              **kwargs)
        self.weights = _check_weights('uniform')

    @property
    def _pairwise(self):
        # For cross-validation routines to split data correctly
        return self.metric == 'precomputed'

    def kneighbors_sim(self,X):
        """
        Find the k-nearest neighbours for each instance and similarity scores. 
        All distances (D) are converted to similarity (S) by:
        
                 D - D.min()
        Sim =   --------------
                D.max()-D.min()
        We assume D.min()==0

        """
        neigh_dist, neigh_ind = self.kneighbors(X)
        
        # Convert distances to similarities:
        if self.metric == 'jaccard':
            neigh_sim = 1-neigh_dist
        else:
            neigh_dist_n = neigh_dist / neigh_dist.max()
            neigh_sim = 1 - neigh_dist_n                       
        
        return neigh_sim, neigh_ind
    
    def predict(self, X):
        """Predict the target for the provided data

        Parameters
        ----------
        X : array-like, shape (n_queries, n_features), \
                or (n_queries, n_indexed) if metric == 'precomputed'
            Test samples.

        Returns
        -------
        y : array of int, shape = [n_queries] or [n_queries, n_outputs]
            Target values
        """
        X = check_array(X, accept_sparse='csr')

        neigh_sim, neigh_ind = self.kneighbors_sim(X)
        
        _y = self._y
        if _y.ndim == 1:
            _y = _y.reshape((-1, 1))

        y_pred = np.empty((X.shape[0], _y.shape[1]), dtype=np.float64)

        for j in range(_y.shape[1]):
            denom=np.sum(neigh_sim[j])
            num = np.sum(_y[neigh_ind, j] * neigh_sim[j], axis=1)
            y_pred[:, j] = num / denom

        if self._y.ndim == 1:
            y_pred = y_pred.ravel()

        return y_pred




from matplotlib.collections import PatchCollection
from textwrap import wrap
import pandas as pd
import math

from genra.utl.ixml import *
from genra.rax.skl.reg import GenRAPredValue
from genra.chm.viz import *

class GenRAView:
  
  def __init__(self,*args,X=None, Y=None, 
               knn=25,kekulize=True,wedgeBonds=True,metric='jaccard',
               dt='c',text_wrap_len=30,chm_name_font_size=8,
               **kwargs):
    self._kekulize = kekulize
    self._wedgeBonds= wedgeBonds
    self._X = X
    self._Y = Y
    self._dt=dt
    self._txt_wrap=text_wrap_len
    self._chm_fnsz=chm_name_font_size
    self._GV = GenRAPredValue(n_neighbors=knn,metric=metric,n_jobs=-1)

  def loadData(self,X,Y=[],Info=None):
    self._X = X
    self._Y = Y
    self._I = Info
    
    I2I = dict(zip(range(X.shape[0]),X.index))
    
    self._GV.fit(X,Y if len(Y)>0 else np.ones(X.shape[0]))
    Sim, Ind = self._GV.kneighbors_sim(self._X)
    
    # Store as dicts for lookup using the index 
    self._NN_sim = dict(zip(X.index,Sim))
    self._NN_ind = dict(zip(X.index,[[I2I[i] for i in I] for I in Ind]))
    
  def getKNN(self,cid,k=10):
    Info = self._I
    Sim,Ind = self._NN_sim[cid],self._NN_ind[cid]    
    NNi= pd.DataFrame(dict(ID=Ind,sim=Sim)).iloc[:(k+1)]
    self._NNi = NNi.merge(self._I,on='ID')
  
class GenRAViewNN(GenRAView):
  def __init__(self,th_tot=math.pi,th0=0.5*math.pi,
                 BBox=(-500,-500,500,500),Origin=[0,0],
                 pred = False,t0=None,r_min=100,rs=10,
                 chm_sz=(100,100),
                 ax = None,
                 use_svg=False,
                 **kwargs):
    GenRAView.__init__(self,**kwargs)
    self._th_tot = th_tot
    self._th0    = th0
    self._BBox   = BBox
    self._O      = Origin
    self._pred   = pred
    self._r_min  = r_min
    self._rs     = rs
    self._chm_sz = chm_sz
    self._ax      = ax
    self._CV = ChemJSView(ax=ax,xyoff=[0,0],fs=1.5,molsize=chm_sz) \
                if use_svg else  ChemView(ax=ax,xyoff=[0,0],fs=1.5,molsize=chm_sz)
    
  def circLayout(self,cid,k=10):    
    dth = self._th_tot/k
    xo,yo = self._O
    xmin,ymin,xmax,ymax= self._BBox
    r_max = abs(xmax-xo)-self._r_min
    #print(r_max)
    self.getKNN(cid,k=k)
    NN_c = self._NNi.iloc[1:]
    NN_c.loc[:,'r'] = self._r_min+r_max*(1-NN_c.sim)*self._rs
    NN_c.loc[:,'th'] = self._th0-(np.arange(0,NN_c.shape[0]) * dth)
    NN_c.loc[:,'x'] = NN_c.r*np.cos(NN_c.th)
    NN_c.loc[:,'y'] = NN_c.r*np.sin(NN_c.th)
    NN_c.loc[:,'lx']= 0.5*(NN_c.x+xo)
    NN_c.loc[:,'ly']= 0.5*(NN_c.y+yo)
    
    self._NN_c = NN_c
    
  def draw(self,cid,k=10,pred=False,t0=None):
    self.circLayout(cid,k=k)
    self.drawArcs()
    self.drawChems()
    
  def drawArcs(self):
    ch_w,ch_h = self._chm_sz
    ch_hw,ch_hh= 0.5*ch_w,0.5*ch_h
    xo,yo = self._O
    dsc_lab= self._dt
    
    ax = self._ax

    for i,C in self._NN_c.iterrows():
      ax.add_patch(FancyArrowPatch(posB=(C.x,C.y),posA=(xo,yo),
                                   color='grey',arrowstyle='->',
                                   #shrinkA=60,shrinkB=60,
                                   linewidth=2,zorder=0))
      ax.add_artist(text.Text(C.lx,C.ly,'{:3.2f}{}'.format(C.sim,dsc_lab),
                              color=(1.0,0.1,0.1),
                              ha='center',va='top',zorder=10,
                              backgroundcolor='white',
                              fontproperties=fm.FontProperties(size=8,weight='normal')))

  def drawChems(self):
    ch_w,ch_h = self._chm_sz
    ch_hw,ch_hh= 0.5*ch_w,0.5*ch_h
    xo,yo = self._O
    CV = self._CV
    ax = self._ax
    
    BGs=[]
    
    # Target
    BGs.append(mpatches.Rectangle((xo-ch_hw,xo-ch_hh),ch_w,ch_h,zorder=1))
    Tgt = self._NNi.iloc[0]
    CV(Tgt.smiles,x0=xo,y0=yo)
    txt = '\n'.join(wrap(Tgt.chemical_name,self._txt_wrap))
    ax.add_artist(text.Text(xo,yo+ch_hh,txt,
                            color=(0.2,0.3,0.2),
                            ha='center',va='top',zorder=10,
                            backgroundcolor='white',
                            fontproperties=fm.FontProperties(size=self._chm_fnsz,
                                                             weight='normal')))

    
    for i,C in self._NN_c.iterrows():
      x0, y0 = C.x-ch_hw,C.y-ch_hh
      BGs.append(mpatches.Rectangle((C.x-ch_hw,C.y-ch_hh),ch_w,ch_h,zorder=1))
      try:
        CV(C.smiles,x0=C.x,y0=C.y)
      except:
        print("Failed: {}".format(C.chemical_name))
      ax.add_artist(text.Text(C.x,C.y+ch_hh,
                              '\n'.join(wrap(C.chemical_name,self._txt_wrap)),
                              color=(0.2,0.3,0.2),
                              ha='center',va='top',zorder=10,
                              backgroundcolor='white',
                              fontproperties=fm.FontProperties(size=self._chm_fnsz,
                                                             weight='normal')))

      
      
    PC = PatchCollection(BGs,alpha=0.0)
    ax.add_collection(PC)
    



import re

from .ixml import *

class Svg2Json:
  def __init__(self,*args,**kwargs):
    self._xmlparser = Xml2Obj()
    self._svgtypes = ['polygon','text','line']  
    self._rx=re.compile(r'\d+')
    self._rxp=re.compile(r'(?P<command>\w+)\s+(?P<x>\d+\.\d+),(?P<y>\d+\.\d+)')
  def __call__(self,svg_str,**kwargs):
    return self.read_svg(svg_str)
    
  def read_svg(self,svg_str,fontsize=6.0,**kwargs):
    J = elemDict(self._xmlparser.Parse(svg_str))
    K = self._svgtypes
    
    #return J
    rx = self._rx
    rxp= self._rxp
   
    if 'text' in J:
      J['text'] = [self.getText(j) for j in J['text']]
      
    if 'line' in J:
      J['line'] = [self.getLine(i) for i in J['line']]

    if 'polygon' in J:
      J['polygon'] = [self.getPolygon(i) for i in J['polygon']]
      
    if 'path' in J:
      J['path'] = [self.getPath(i) for i in J['path']]

    return J

  def getText(self,In):
    Out = {}
    Out.update({k.replace('-','_'):self.fix(k,v) for k,v in In.items()})
    Out['text'] = Out.pop('tspan')
    if 'style' in Out: Out.update(self.getStyles(Out['style']))
      
    return Out
 
  def getLine(self,In):
    Out = {}
    Out.update({k.replace('-','_'):self.fix(k,v) for k,v in In.items()})
    
    return Out
  
  def getPolygon(self,In):
    Out = {}
    Out.update({k.replace('-','_'):self.fix(k,v) for k,v in In.items()})
    
    return Out
  
  def getPath(self,In):
    Out = {}
    Out.update({k:self.fix(k,v) for k,v in In.items()})
    Out['lines'] = [m.groupdict() for m in self._rxp.finditer(Out['d'])]
    if 'style' in Out: Out.update(self.getStyles(Out['style']))

    return Out
 
  def fixpoly(self,x):
    p = rxp.findall(x)
    if not p: return x
    P = "M%s Z" % " L".join(p)
    return P

  def fixrgb(self,x):
    a = rx.findall(x)
    if len(a) != 3:
      return "#ffffff"
    else:
      return "#%(r)02X%(g)02X%(b)02X" % dict(zip(('r','g','b'),map(int,a)))

  def fix(self,k,v):
    if k=='tspan':
      v = v[0]['cdata']
    elif k in ['x','y','font-size']:
      v = float(v)
      try:
        v = float(v)
      except:
        pass
    return v
      
  def getStyles(self,x):
    Y = {k:v for k,v in [i.split(':') for i in x.split(';')]}
    for atr in ['font-size','stroke-width','stroke-opacity','fill-opacity']:
      if atr in Y: 
        val = Y[atr]
        try:
          val = float(re.sub('\D+$','',val))
        except:
          pass
        else:
          Y[atr]=val
    return Y
      
  def old(self,x):
    if 'fill' in x:
      x['fill']=self.fixrgb(x['fill'])
    if 'color' in x:
      x['color']=self.fixrgb(x['color'])
    if 'x' in x:
      x['x']=float(x['x'])
    if 'y' in x:
      x['y']=float(x['y'])
    if 'points' in x:
      x['points']=self.fixpoly(x['points'])
    if 'font-size' in x:
      try:
        x['font-size'] = float(x['font-size'])
      except:
        x['font-size'] = fontsize
    return x

    


"""
GenRA Classifier

Adapted from sklearn.neighbors.KNeighborsClassifier

"""

# Authors: Imran Shah (shah.imran@epa.gov)

import numpy as np
import sklearn
from sklearn.neighbors import KNeighborsClassifier
from sklearn.neighbors._base import _check_weights, _get_weights
from sklearn.utils.validation import check_X_y, check_array, check_is_fitted,_is_arraylike, _num_samples
from sklearn.utils.multiclass import unique_labels
from scipy import stats
from sklearn.utils.extmath import weighted_mode

import warnings

class GenRAPredClass(KNeighborsClassifier):
    """GenRA Classifier implementing the k-nearest neighbors vote.

    Parameters
    ----------
    n_neighbors : int, optional (default = 5)
        Number of neighbors to use by default for :meth:`kneighbors` queries.

    algorithm : {'auto', 'ball_tree', 'kd_tree', 'brute'}, optional
        Algorithm used to compute the nearest neighbors:

        - 'ball_tree' will use :class:`BallTree`
        - 'kd_tree' will use :class:`KDTree`
        - 'brute' will use a brute-force search.
        - 'auto' will attempt to decide the most appropriate algorithm
          based on the values passed to :meth:`fit` method.

        Note: fitting on sparse input will override the setting of
        this parameter, using brute force.

    leaf_size : int, optional (default = 30)
        Leaf size passed to BallTree or KDTree.  This can affect the
        speed of the construction and query, as well as the memory
        required to store the tree.  The optimal value depends on the
        nature of the problem.

    metric : string or callable, default '1-jaccard'
        the distance metric to use for the tree.  The default metric is
        1-jaccard, and with p=2 is equivalent to the standard Euclidean
        metric. See the documentation of the DistanceMetric class for a
        list of available metrics.

    metric_params : dict, optional (default = None)
        Additional keyword arguments for the metric function.

    n_jobs : int
    or None, optional (default=None)
        The number of parallel jobs to run for neighbors search.
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
        ``-1`` means using all processors. See :term:`Glossary <n_jobs>`
        for more details.
        Doesn't affect :meth:`fit` method.

    Attributes
    ----------
    classes_ : array of shape (n_classes,)
        Class labels known to the classifier

    effective_metric_ : string or callble
        The distance metric used. It will be same as the `metric` parameter
        or a synonym of it, e.g. 'euclidean' if the `metric` parameter set to
        'minkowski' and `p` parameter set to 2.

    effective_metric_params_ : dict
        Additional keyword arguments for the metric function. For most metrics
        will be same with `metric_params` parameter, but may also contain the
        `p` parameter value if the `effective_metric_` attribute is set to
        'minkowski'.

    outputs_2d_ : bool
        False when `y`'s shape is (n_samples, ) or (n_samples, 1) during fit
        otherwise True.

    """

    def __init__(self, n_neighbors=5,
                 weights='uniform', algorithm='auto', leaf_size=30,
                 p=2, metric='jaccard', metric_params=None, n_jobs=None,
                 **kwargs):
        super().__init__(
            n_neighbors=n_neighbors,
            algorithm=algorithm,
            leaf_size=leaf_size, metric=metric, p=p,
            metric_params=metric_params,
            n_jobs=n_jobs, **kwargs)
        self.weights = _check_weights(weights)

    def predict(self, X):
        """Predict the class labels for the provided data.

        Parameters
        ----------
        X : array-like, shape (n_queries, n_features), \
                or (n_queries, n_indexed) if metric == 'precomputed'
            Test samples.

        Returns
        -------
        y : array of shape [n_queries] or [n_queries, n_outputs]
            Class labels for each data sample.
        """
        X = check_array(X, accept_sparse='csr')

        neigh_dist, neigh_ind = self.kneighbors(X)

        classes_ = self.classes_
        _y = self._y
        if not self.outputs_2d_:
            _y = self._y.reshape((-1, 1))
            classes_ = [self.classes_]

        n_outputs = len(classes_)
        n_queries = _num_samples(X)
        weights = _get_weights(neigh_dist, self.weights)

        y_pred = np.empty((n_queries, n_outputs), dtype=classes_[0].dtype)
        for k, classes_k in enumerate(classes_):
            if weights is None:
                mode, _ = stats.mode(_y[neigh_ind, k], axis=1)
            else:
                mode, _ = weighted_mode(_y[neigh_ind, k], weights, axis=1)

            mode = np.asarray(mode.ravel(), dtype=np.intp)
            y_pred[:, k] = classes_k.take(mode)

        if not self.outputs_2d_:
            y_pred = y_pred.ravel()

        return y_pred
        
    
    def predict_proba(self, X):
        """Return probability estimates for the test data X.

        Parameters
        ----------
        X : array-like, shape (n_queries, n_features), \
                or (n_queries, n_indexed) if metric == 'precomputed'
            Test samples.

        Returns
        -------
        p : array of shape = [n_queries, n_classes], or a list of n_outputs
            of such arrays if n_outputs > 1.
            The class probabilities of the input samples. Classes are ordered
            by lexicographic order.
        """
        X = check_array(X, accept_sparse='csr')

        neigh_dist, neigh_ind = self.kneighbors(X)

        classes_ = self.classes_
        _y = self._y
        if not self.outputs_2d_:
            _y = self._y.reshape((-1, 1))
            classes_ = [self.classes_]

        n_queries = _num_samples(X)

        weights = _get_weights(neigh_dist, self.weights)
        if weights is None:
            weights = np.ones_like(neigh_ind)

        all_rows = np.arange(X.shape[0])
        probabilities = []
        for k, classes_k in enumerate(classes_):
            pred_labels = _y[:, k][neigh_ind]
            proba_k = np.zeros((n_queries, classes_k.size))

            # a simple ':' index doesn't work right
            for i, idx in enumerate(pred_labels.T):  # loop is O(n_neighbors)
                proba_k[all_rows, idx] += weights[:, i]

            # normalize 'votes' into real [0,1] probabilities
            normalizer = proba_k.sum(axis=1)[:, np.newaxis]
            normalizer[normalizer == 0.0] = 1.0
            proba_k /= normalizer

            probabilities.append(proba_k)

        if not self.outputs_2d_:
            probabilities = probabilities[0]

        return probabilities




from setuptools import find_packages, setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


setup(
    name='genra',
    packages=find_packages(),
    version='0.2',
    install_requires=['numpy', 'scipy', 'scikit-learn==1.3.1', 'lxml'],
    description='Generalised Read Across (GenRA) in Python',
    long_description = long_description,
    long_description_content_type = "text/markdown",
    url = "https://github.com/i-shah/genra-py",
    author='Imran Shah',
    license='MIT',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3'],
        keywords = 'genra'
)
